import pandas as pd
import geopandas as gpd
import json

from sqlalchemy import text
from galigeopy.model.properties.property import Property
from galigeopy.utils.types import pythonTypeToPostgresType

class GeolevelDataType:
    # Constructor
    def __init__(
            self,
            geoleveldata_type_id:int,
            name:str,
            properties:dict,
            geolevel_id:int,
            description:str,
            org:'Org' # type: ignore
    ):
        self._geoleveldata_type_id = geoleveldata_type_id
        self._name = name
        self._properties = properties
        self._geolevel_id = geolevel_id
        self._description = description
        self._org = org

    # Getters
    @property
    def geoleveldata_type_id(self): return self._geoleveldata_type_id
    @property
    def name(self): return self._name
    @property
    def properties(self): return self._properties
    @property
    def geolevel_id(self): return self._geolevel_id
    @property
    def description(self): return self._description
    @property
    def org(self): return self._org

    # Setters
    @name.setter
    def name(self, value): self._name = value
    @description.setter
    def description(self, value): self._description = value

    # Public Methods
    def getGeolevel(self):
        return self._org.getGeolevelById(self._geolevel_id)
    
    def getDataset(self)->pd.DataFrame:
        query = text(f"SELECT * FROM ggo_geoleveldata WHERE geoleveldata_type_id = {self._geoleveldata_type_id}")
        return pd.read_sql(query, self._org.engine)
    
    def getGeoDataset(self, columns=None)->gpd.GeoDataFrame:
        geolevel = self.getGeolevel()
        df_columns = pd.DataFrame(self._properties["columns"])
        if columns is not None:
            df_columns = df_columns[df_columns["colonne"].isin(columns)]
        p = []
        for index, row in df_columns.iterrows():
            p.append(f"(g.properties->>'{row['colonne']}')::{row['type']} AS {row['colonne']}")
        properties = ", ".join(p)
        properties = properties + "," if properties != "" else ""
        query = text(f"""
            SELECT
                g.geounit_code,
                {properties}
                geo.{geolevel.geom_field} AS geometry
            FROM ggo_geoleveldata g
            JOIN {geolevel.table_name} AS geo ON geo.{geolevel.geounit_code} = g.geounit_code
            WHERE
                geoleveldata_type_id = {self._geoleveldata_type_id}
        """)
        return gpd.read_postgis(query, self._org.engine, geom_col='geometry')
    
    def add_to_model(self) -> 'GeolevelDataType':
        # add to database
        query = f"""
            INSERT INTO ggo_geoleveldata_type (
                name,
                description,
                properties,
                geolevel_id
            ) VALUES (
                '{self._name.replace("'", "''")}',
                '{self._description.replace("'", "''")}',
                '{json.dumps(self._properties).replace("'", "''")}',
                {self._geolevel_id}
            ) RETURNING geoleveldata_type_id
        """
        geolevel_data_type_id = self._org.query(query)[0][0]
        self._geoleveldata_type_id = geolevel_data_type_id
        # return
        return self
    
    def update(self) -> 'GeolevelDataType':
        query = f"""
            UPDATE ggo_geoleveldata_type
            SET
                name = '{self._name.replace("'", "''")}',
                description = '{self._description.replace("'", "''")}',
                properties = '{json.dumps(self._properties).replace("'", "''")}',
                geolevel_id = {self._geolevel_id}
            WHERE geoleveldata_type_id = {self._geoleveldata_type_id}
        """
        self._org.query(query)
        return self
    
    def delete(self):
        query = f"DELETE FROM ggo_geoleveldata_type WHERE geoleveldata_type_id = {self._geoleveldata_type_id}"
        self._org.query(query)
        self._geoleveldata_type_id = None
        return True

    def getProperties(self, fast:int|None=None)->list[Property]:
        query = f"""
            SELECT
                gd.properties
            FROM ggo_geoleveldata AS gd
            WHERE gd.geoleveldata_type_id = {self._geoleveldata_type_id}
        """
        query += f" LIMIT {fast}" if fast is not None else ""
        list_properties = self._org.query_df(query)['properties'].tolist()
        df_properties = pd.DataFrame(list_properties)
        df_prop = df_properties.dtypes.reset_index()
        df_prop.columns = ['columns', 'dtypes']
        df_prop['dtypes_postgres'] = df_prop['dtypes'].astype(str).apply(lambda x: pythonTypeToPostgresType(x))
        p = []
        for index, row in df_prop.iterrows():
            prop = Property(column='properties', dtype='JSONB', json_info={"key": row['columns'], "dtype": row['dtypes_postgres']})
            p.append(prop)
        return p

    
    @staticmethod
    def getAllGeolevelDataTypesOfGeolevel(geolevel):
        q = f"""
            SELECT
                DISTINCT gdt.*
            FROM ggo_geoleveldata_type AS gdt
            WHERE gdt.geolevel_id = {geolevel.geolevel_id}
        """
        org = geolevel.org
        df = geolevel.org.query_df(q)
        return [GeolevelDataType(**row, org=org) for row in df.to_dict(orient="records")]



import requests
import asyncio
import aiohttp
import json
from pydantic import BaseModel
from typing import Optional, List

from .engine import Engine

class OsrmEngineParameters(BaseModel):
    url: str
    verified_url: bool
    default_profile: Optional[str] = "driving"
    default_version: Optional[str] = "v1"

class OsrmEngine(Engine):
    def __init__(self, parameters: dict, auto_test: bool = True):
        self._url = parameters.get('url')
        self._verified_url = parameters.get('verified_url')
        self._default_profile = parameters.get('default_profile', 'driving')
        self._default_version = parameters.get('default_version', 'v1')
        super().__init__("osrm", parameters, OsrmEngineParameters, auto_test=auto_test)

    @property
    def url(self): return self._url
    @property
    def verified_url(self): return self._verified_url
    @property
    def default_profile(self): return self._default_profile
    @property
    def default_version(self): return self._default_version

    def _check_status(self):
        """
        Check if the OSRM server is reachable.
        """
        try:
            response = requests.get(f"{self._url}route/v1/driving/13.388860,52.517037;13.397634,52.529407?overview=false", verify=self._verified_url)
            if response.status_code == 200:
                return True
            else:
                raise Exception(f"OSRM server is not reachable: {response.status_code} - {response.text}")
        except requests.RequestException as e:
            raise Exception(f"Error connecting to OSRM server: {e}")
    
    def list_functions(self):
        return ["get_nearest", "get_route", "get_route_async", "get_table", "get_match", "get_trip", "get_tile"]
    
    def get_nearest(self, location:dict, number:int=1, version:str=None, profile:str=None)->list:
        """
        Get the nearest waypoints to a location.
        :param location: A dictionary with 'lat' and 'lng' keys.
        :param number: The number of nearest waypoints to return.
        :return: A list of waypoints.
        """
        v = version if version else self.default_version
        p = profile if profile else self.default_profile
        url = f"{self.url.removesuffix('/')}/nearest/{v}/{p}/{location['lng']},{location['lat']}?number={number}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        return data["waypoints"]
    
    def get_route(self, start, end, version:str=None, profile:str=None)->list:
        """
        Get the route between two locations.
        :param start: A dictionary with 'lat' and 'lng' keys for the start location.
        :param end: A dictionary with 'lat' and 'lng' keys for the end location.
        :return: A list of routes.
        """
        v = version if version else self.default_version
        p = profile if profile else self.default_profile
        url = f"{self.url.removesuffix('/')}/route/{v}/{p}/{start['lng']},{start['lat']};{end['lng']},{end['lat']}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code == 200:
            data = response.json()
            if data["code"] == "Ok":
                return data["routes"]
            else:
                return []
        elif response.status_code == 400:
            data = response.json()
            if data["code"] == "NoRoute":
                return []
            else:
                raise Exception(f"Error {data['code']}: {data['message']}")
        else:
            raise Exception(f"Error {response.status_code}: {response.text}")
        
    def get_route_async(
        self,
        start:list,
        end:list,
        version:str=None,
        profile:str=None,
        alternatives:bool=False,
        steps:bool=False,
        annotations:bool=False,
        geometries:str="polyline",
        overview:str="simplified",
        continue_straight:str="default"
    )->list:
        """
        Get the route between multiple pairs of locations asynchronously.
        :param start: A list of dictionaries with 'lat' and 'lng' keys for the start locations.
        :param end: A list of dictionaries with 'lat' and 'lng' keys for the end locations.
        :return: A list of routes.
        """
        # Check start and end have the same length
        if len(start) != len(end):
            raise Exception("Start and end must have the same length")
        # Prepare urls
        v = version if version else self.default_version
        p = profile if profile else self.default_profile
        urls = []
        for i in range(len(start)):
            url = f"{self.url.removesuffix('/')}/route/{v}/{p}/{start[i]['lng']},{start[i]['lat']};{end[i]['lng']},{end[i]['lat']}"
            # Properties
            url += f"?alternatives={str(alternatives).lower()}"
            url += f"&steps={str(steps).lower()}"
            url += f"&annotations={str(annotations).lower()}"
            url += f"&geometries={geometries}"
            url += f"&overview={overview}"
            url += f"&continue_straight={continue_straight}"
            urls.append(url)
        # Async
        async def get(url, session):
            try:
                async with session.get(url=url) as response:
                    resp = await response.read()
                    return resp
            except Exception as e:
                pass
        async def main(urls):
            async with aiohttp.ClientSession() as session:
                ret = await asyncio.gather(*(get(url, session) for url in urls))
                return ret
        # Run
        data = asyncio.run(main(urls))
        # Check if noRoute
        json_data = [json.loads(d) for d in data]
        return [d for d in json_data]
    
    def get_table(self, locations:list, sources:list=None, destinations:list=None, version:str=None, profile:str=None)->list:
        """
        Get the distance and duration table between multiple locations.
        :param locations: A list of dictionaries with 'lat' and 'lng' keys for the locations.
        :param sources: A list of indices for the source locations.
        :param destinations: A list of indices for the destination locations.
        :return: A dictionary with the distance and duration table.
        """
        v = version if version else self.default_version
        p = profile if profile else self.default_profile
        url = f"{self.url.removesuffix('/')}/table/{v}/{p}/"
        url += f";".join([f"{location['lng']},{location['lat']}" for location in locations])
        if sources or destinations:
            url += "?"
        if sources:
            url += f"sources={';'.join([str(source) for source in sources])}"
        if destinations:
            if sources:
                url += "&"
            url += f"destinations={';'.join([str(destination) for destination in destinations])}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        del data["code"]
        return data
    
    def get_match(
        self,
        locations:list,
        steps:bool=False,
        geometries:str="polyline",
        annotations:bool=False,
        overview:str="simplified",
        timestamps:list=None,
        radiuses:list=None,
        version:str=None,
        profile:str=None
    )->list:
        """
        Match a GPS trace to the road network.
        :param locations: A list of dictionaries with 'lat' and 'lng' keys for the locations.
        :param steps: Whether to include steps in the response.
        :param geometries: The format of the geometries to return.
        :param annotations: Whether to include annotations in the response.
        :param overview: The level of detail for the overview geometry.
        :param timestamps: A list of timestamps for the locations.
        :param radiuses: A list of radiuses for the locations.
        :return: A list of matched routes.
        """
        v = version if version else self.default_version
        p = profile if profile else self.default_profile
        url = f"{self.url.removesuffix('/')}/match/{v}/{p}/"
        url += f";".join([f"{location['lng']},{location['lat']}" for location in locations])
        # Properties
        url += f"?steps={str(steps).lower()}"
        url += f"&geometries={geometries}"
        url += f"&annotations={str(annotations).lower()}"
        url += f"&overview={overview}"
        if timestamps:
            url += f"&timestamps={';'.join([str(timestamp) for timestamp in timestamps])}"
        if radiuses:
            url += f"&radiuses={';'.join([str(radius) for radius in radiuses])}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        del data["code"]
        return data
    
    def get_trip(
        self,
        locations:list,
        steps:bool=False,
        annotations:bool=False,
        geometries:str="polyline",
        overview:str="simplified",
        version:str=None,
        profile:str=None
    ):
        """
        Get a trip between multiple locations.
        :param locations: A list of dictionaries with 'lat' and 'lng' keys for the locations.
        :param steps: Whether to include steps in the response.
        :param annotations: Whether to include annotations in the response.
        :param geometries: The format of the geometries to return.
        :param overview: The level of detail for the overview geometry.
        :return: A dictionary with the trip information.
        """
        v = version if version else self.default_version
        p = profile if profile else self.default_profile
        url = f"{self.url.removesuffix('/')}/trip/{v}/{p}/"
        url += f";".join([f"{location['lng']},{location['lat']}" for location in locations])
        # Properties
        url += f"?steps={str(steps).lower()}"
        url += f"&annotations={str(annotations).lower()}"
        url += f"&geometries={geometries}"
        url += f"&overview={overview}"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        data = response.json()
        if data["code"] != "Ok":
            raise Exception(f"Error {data['code']}: {data['message']}")
        del data["code"]
        return data
    
    def get_tile(
        self,
        x:int,
        y:int,
        zoom:int,
        version:str=None,
        profile:str=None
    )->bytes:
        """
        Get a tile from the OSRM server.
        :param x: The x coordinate of the tile.
        :param y: The y coordinate of the tile.
        :param zoom: The zoom level of the tile.
        :return: The tile data as bytes.
        """
        v = version if version else self.default_version
        p = profile if profile else self.default_profile
        url = f"{self.url.removesuffix('/')}/tile/{v}/{p}/tile({x},{y},{zoom}).mvt"
        response = requests.get(url, verify=self.verified_url)
        if response.status_code != 200:
            raise Exception(f"Error {response.status_code}: {response.text}")
        return response.content


import pandas as pd
import geopandas as gpd
from sqlalchemy import text
import json

from galigeopy.model.zone import Zone
from galigeopy.model.poi import Poi
from galigeopy.utils.types import pythonTypeToPostgresType
from galigeopy.model.properties.property import Property

class ZoneType:
    def __init__(
        self,
        zone_type_id:int,
        name:str,
        description:str,
        org: 'Org' # type: ignore
    ):
        # Infos
        self._zone_type_id = zone_type_id
        self._name = name
        self._description = description
        # Engine
        self._org = org

    def __str__(self):
        return self.name + " (" + str(self.zone_type_id) + ")"
    
    # Getters
    @property
    def zone_type_id(self): return self._zone_type_id
    @property
    def name(self): return self._name
    @property
    def description(self): return self._description
    @property
    def org(self): return self._org

    # Setters
    @name.setter
    def name(self, value): self._name = value
    @description.setter
    def description(self, value): self._description = value
    
    # Public Methods
    def number_of_zones(self):
        query = text(f"SELECT COUNT(*) FROM ggo_zone WHERE zone_type_id = {self.zone_type_id}")
        with self._org.engine.connect() as conn:
            result = conn.execute(query)
            return result.scalar()
        
    def getZonesList(self):
        query = text(f"SELECT * FROM ggo_zone WHERE zone_type_id = {self.zone_type_id}")
        df = self._org.query_df(query)
        # Properties to json
        df['properties'] = df['properties'].apply(lambda x: json.loads(x) if isinstance(x, str) else x)
        return df

    def getZoneById(self, zone_id:int):
        query = f"SELECT * FROM ggo_zone WHERE zone_id = {zone_id} AND zone_type_id = {self.zone_type_id}"
        df = self._org.query_df(query)
        if len(df) > 0:
            data = df.iloc[0].to_dict()
            data.update({"org": self._org})
            return Zone(**data) 
        else:
            raise Warning(f"Zone {zone_id} not found in ZoneType {self.name}")
        
    def getZonesByPoi(self, poi : Poi):
        query = text(f"SELECT * FROM ggo_zone WHERE poi_id = {poi.poi_id} AND zone_type_id = {self.zone_type_id}")
        r = pd.read_sql(query, self._org.engine)
        zones = [Zone(**data, org=self._org) for data in r.to_dict(orient='records')]
        return zones
        
    def getAllZones(self):
        query = text(f"SELECT * FROM ggo_zone WHERE zone_type_id = {self.zone_type_id}")
        gdf = gpd.read_postgis(query, self._org.engine, geom_col='geometry')
        zones = []
        for i in range(len(gdf)):
            data = gdf.iloc[i].to_dict()
            data.update({"org": self._org})
            zones.append(Zone(**data))
        return zones
    
    def getAllPois(self) -> list:
        query = text(f"SELECT DISTINCT p.* FROM ggo_poi AS p JOIN ggo_zone AS z ON p.poi_id = z.poi_id WHERE z.zone_type_id = {self.zone_type_id} ORDER BY p.poi_id")
        gdf = gpd.read_postgis(query, self._org.engine, geom_col='geom')
        pois = [Poi(**data, org=self._org) for data in gdf.to_dict(orient='records')]
        # pois = list(set(pois))
        return pois
        
    def add_to_model(self)-> int:
        # Add to database
        query = f"""
        INSERT INTO ggo_zone_type (
            name,
            description
        ) VALUES (
            '{self.name.replace("'", "''")}',
            '{self.description.replace("'", "''")}'
        ) RETURNING zone_type_id;
        """
        zone_type_id = self._org.query(query)[0][0]
        return zone_type_id
    
    def update(self) -> 'ZoneType':
        query = f"""
        UPDATE ggo_zone_type
        SET
            name = '{self.name.replace("'", "''")}',
            description = '{self.description.replace("'", "''")}'
        WHERE zone_type_id = {self.zone_type_id}
        """
        self._org.query(query)
        return self._org.getZoneTypeById(self.zone_type_id)
    
    def delete(self)-> bool:
        query = f"DELETE FROM ggo_zone_type WHERE zone_type_id = {self.zone_type_id}"
        self._org.query(query)
        self._zone_type_id = None
        return True

    def getZonesProperties(self, fast:int|None=None)-> list[Property]:
        # Other properties
        query = f"""
        SELECT
            z.properties
        FROM ggo_zone AS z
        WHERE z.zone_type_id = {self.zone_type_id};
        """
        query += f" LIMIT {fast}" if fast else ""
        list_properties = self._org.query_df(query)['properties'].tolist()
        df_properties = pd.DataFrame(list_properties)
        df_prop = df_properties.dtypes.reset_index()
        df_prop.columns = ['columns', 'dtypes']
        df_prop['dtypes_postgres'] = df_prop['dtypes'].astype(str).apply(lambda x: pythonTypeToPostgresType(x))
        p = []
        for index, row in df_prop.iterrows():
            prop = Property(column='properties', dtype='JSONB', json_info={"key": row['columns'], "dtype": row['dtypes_postgres']})
            p.append(prop)
        return p
    
    def getZonesGeounitsProperties(self, fast:int|None=None)-> list[Property]:
        query = f"""
        SELECT
            zg.properties
        FROM ggo_zone_geounit AS zg
        JOIN ggo_zone AS z ON z.zone_id = zg.zone_id
        WHERE z.zone_type_id = {self.zone_type_id};
        """
        query += f" LIMIT {fast}" if fast else ""
        list_properties = self._org.query_df(query)['properties'].tolist()
        df_properties = pd.DataFrame(list_properties)
        df_prop = df_properties.dtypes.reset_index()
        df_prop.columns = ['columns', 'dtypes']
        df_prop['dtypes_postgres'] = df_prop['dtypes'].astype(str).apply(lambda x: pythonTypeToPostgresType(x))
        p = []
        for index, row in df_prop.iterrows():
            prop = Property(column='properties', dtype='JSONB', json_info={"key": row['columns'], "dtype": row['dtypes_postgres']})
            p.append(prop)
        return p
    
    def getProperties(self, fast:int|None=None)-> list[Property]:
        return self.getZonesProperties(fast=fast) + self.getZonesGeounitsProperties(fast=fast)
    
    @staticmethod
    def getAllZoneTypesOfNetwork(network):
        query = f"""
        SELECT
            DISTINCT zt.*
        FROM ggo_poi AS p
        JOIN ggo_zone AS z ON p.poi_id = z.poi_id
        JOIN ggo_zone_type AS zt ON z.zone_type_id = zt.zone_type_id
        WHERE p.network_id = {network.network_id}
        """
        org = network.org
        df = org.query_df(query)
        zone_types = [ZoneType(**data, org=org) for data in df.to_dict(orient='records')]
        return zone_types


import pandas as pd

class ClusterSession:
    def __init__(
        self,
        cluster_session_id:int | None,
        name:str,
        description:str,
        network_id:int | None,
        geolevel_id:int | None,
        properties:dict,
        org: 'Org' # type: ignore
    ):
        # Infos
        self._cluster_session_id = cluster_session_id
        self._name = name
        self._description = description
        self._network_id = network_id
        self._geolevel_id = geolevel_id
        self._properties = properties
        # Engine
        self._org = org

    def __str__(self):
        return self.name + " (" + str(self.cluster_session_id) + ")"
    
    # Getters
    @property
    def cluster_session_id(self): return self._cluster_session_id
    @property
    def name(self): return self._name
    @property
    def description(self): return self._description
    @property
    def network_id(self): return self._network_id
    @property
    def geolevel_id(self): return self._geolevel_id
    @property
    def properties(self): return self._properties
    @property
    def org(self): return self._org

    # Setters
    @name.setter
    def name(self, value): self._name = value
    @description.setter
    def description(self, value): self._description = value

    # Public Methods
    def number_of_clusters(self):
        query = f"SELECT COUNT(*) FROM ggo_cluster WHERE cluster_session_id = {self.cluster_session_id}"
        with self._org.engine.connect() as conn:
            result = conn.execute(query)
            return result.scalar()
        
    def getClustersList(self):
        query = f"SELECT * FROM ggo_cluster WHERE cluster_session_id = {self.cluster_session_id}"
        return pd.read_sql(query, self._org.engine)
    
    


class Property:
    def __init__(self, column, dtype=str, json_info: dict[str, str] | None= None, rename=None, agg_function:str|None=None):
        self._column = column
        self._dtype = dtype
        self._json_info = json_info
        self._rename = rename
        self._agg_function = agg_function

    @property
    def column(self): return self._column
    @property
    def dtype(self): return self._dtype
    @property
    def json_info(self): return self._json_info
    @property
    def rename(self): return self._rename
    @property
    def agg_function(self): return self._agg_function

    # setters
    @column.setter
    def column(self, column): self._column = column
    @dtype.setter
    def dtype(self, dtype): self._dtype = dtype
    @json_info.setter
    def json_info(self, json_info): self._json_info = json_info
    @rename.setter
    def rename(self, rename): self._rename = rename
    @agg_function.setter
    def agg_function(self, agg_function): self._agg_function = agg_function

    def to_sql(self, prefix=None):
        # Build the base column reference, possibly with a prefix (e.g. table alias)
        col_ref = f"{prefix}.{self._column}" if prefix else self._column

        # If json_info is provided, extract the key from the JSON column
        if self._json_info:
            json_key = self._json_info["key"]
            json_dtype = self._json_info.get("dtype", "text")
            col_ref = f"({col_ref} ->> '{json_key}')::{json_dtype}"

        # Apply aggregation function if specified
        if self._agg_function:
            col_ref = f"{self._agg_function.upper()}({col_ref})"

        # Apply aliasing: use the rename if provided, otherwise fallback to column or json key
        alias = self.get_name()

        return f"{col_ref} AS {alias}"
    
    def get_name(self):
        return self._rename or self._json_info.get("key", 'undefined') if self._json_info else self._column
    
    def get_type(self):
        return self._dtype if self._dtype.lower() != "jsonb" else self._json_info.get("dtype", "text")
    
    def getDefaultAggFunction(self):
        return "SUM"
    
    def getAllAggFunctions(self):
        return ["SUM", "AVG", "COUNT", "MIN", "MAX"]


from .property import Property
from ..zone_type import ZoneType
import sqlparse

class ZoneTypeProperties:
    def __init__(
            self,
            zone_type: ZoneType,
            basic_zone_properties: list[Property],
            basic_zone_geounit_properties: list[Property],
            network_properties: list['NetworkProperties'] | None = None, # type: ignore
            geolevel_properties: list['GeolevelProperties'] | None = None, # type: ignore
        ):
        self._zone_type = zone_type
        self._basic_zone_properties = basic_zone_properties
        self._basic_zone_geounit_properties = basic_zone_geounit_properties
        self._network_properties = network_properties if network_properties else []
        self._geolevel_properties = geolevel_properties if geolevel_properties else []

    @property
    def zone_type(self): return self._zone_type
    @property
    def basic_zone_properties(self): return self._basic_zone_properties
    @property
    def basic_zone_geounit_properties(self): return self._basic_zone_geounit_properties
    @property
    def network_properties(self): return self._network_properties
    @property
    def geolevel_properties(self): return self._geolevel_properties
    @property
    def properties(self): 
        network_properties = [] 
        geolevel_properties = []
        for np in self._network_properties:
            network_properties += np.properties
        for gp in self._geolevel_properties:
            geolevel_properties += gp.properties
        return self._basic_zone_properties + self._basic_zone_geounit_properties + network_properties + geolevel_properties
    
    def nb_properties(self):
        return len(self._basic_zone_properties) + len(self._basic_zone_geounit_properties) + sum([np.nb_properties() for np in self._network_properties]) + sum([gp.nb_properties() for gp in self._geolevel_properties])

    def sql_basic_zone_properties(self):
        list_sql = []
        if len(self._basic_zone_properties) == 0:
            return "z.*"
        for prop in self._basic_zone_properties:
            list_sql.append(f"{prop.to_sql(prefix='z')}")
        return ", ".join(list_sql)

    def sql_basic_zone_geounit_properties(self):
        list_sql = []
        if len(self._basic_zone_geounit_properties) == 0:
            return "zg.*"
        for prop in self._basic_zone_geounit_properties:
            list_sql.append(f"{prop.to_sql(prefix='zg')}")
        return ", ".join(list_sql)
    
    def properties_sql(self):
        properties = []
        for prop in self._basic_zone_properties:
            properties.append(prop.to_sql(prefix="z"))
        for prop in self._basic_zone_geounit_properties:
            properties.append(prop.to_sql(prefix="zg"))
        for np in self._network_properties:
            properties += np.properties_sql()
        for gp in self._geolevel_properties:
            properties += gp.properties_sql()
        return properties
    
    def join_sql(self, label=False):
        join = ["JOIN ggo_zone_geounit AS zg ON zg.zone_id = z.zone_id"]
        join += ["JOIN ggo_poi AS p ON p.poi_id = z.poi_id"] if label else []
        for np in self._network_properties:
            join += ["JOIN ggo_poi AS p ON p.poi_id = z.poi_id"] if not label else []
            join += np.join_sql()
        for gp in self._geolevel_properties:
            join += ["JOIN ggo_geolevel AS g ON g.geolevel_id = z.geolevel_id"]
            join += gp.join_sql()
        return join

    def where_sql(self):
        where = [f"z.zone_type_id = {self._zone_type.zone_type_id}"]
        for np in self._network_properties:
            where += np.where_sql()
        for gp in self._geolevel_properties:
            where += gp.where_sql()
        return where
    
    def to_sql(self, z_idx=True, zg_idx=True, code=False, label=False):
        properties = [f"z.zone_id AS zone_id"] if z_idx else []
        properties += [f"zg.zone_geounit_id AS zone_geounit_id"] if zg_idx else []
        properties += [f"p.name AS poi_name"] if label else []
        properties += [f"zg.geounit_code AS geounit_code"] if code else []
        properties += self.properties_sql()
        q = f"""
        SELECT
            {', '.join(properties)}
        FROM ggo_zone AS z
        {' '.join(self.join_sql(label=label))}
        WHERE {" AND ".join(self.where_sql())}
        """
        return sqlparse.format(q, reindent=True, keyword_case='upper')


from pydantic import BaseModel

class Engine:
    def __init__(self, engine_type:str,  parameters: dict, base_model: BaseModel, auto_test:bool=True):
        self._engine_type = engine_type
        self._parameters = parameters
        self._base_model = base_model
        self._check_parameters()
        if auto_test:
            self._check_status()

    @property
    def engine_type(self): return self._engine_type
    @property
    def parameters(self): return self._parameters

    def _check_parameters(self):
        try:
            self._base_model(**self._parameters)
        except Exception as e:
            raise ValueError(f"Invalid parameters: {e}")
        
    def list_funtions(self):
        raise NotImplementedError("This method should be implemented in subclasses.")

    def run_function(self, function_name: str, *args, **kwargs):
        if not hasattr(self, function_name):
            raise AttributeError(f"Function '{function_name}' is not defined in the engine.")
        return getattr(self, function_name)(*args, **kwargs)
    
    def _check_status(self):
        raise NotImplementedError("This method should be implemented in subclasses to check the engine status.")
    
        


import sqlparse

from .property import Property
from .zone_type_properties import ZoneTypeProperties
from .distancier_properties import DistancierProperties
from ..network import Network

class NetworkProperties:
    def __init__(self, network: Network, basic_network_properties: list[Property], zone_type_properties: list[ZoneTypeProperties], distancier_properties: list[DistancierProperties]):
        self._network = network
        self._basic_network_properties = basic_network_properties
        self._zone_type_properties = zone_type_properties
        self._distancier_properties = distancier_properties

    @property
    def network(self): return self._network
    @property
    def basic_network_properties(self): return self._basic_network_properties
    @property
    def zone_type_properties(self): return self._zone_type_properties
    @property
    def distancier_properties(self): return self._distancier_properties
    @property
    def properties(self):
        zone_type_properties = []
        distancier_properties = []
        for zt in self._zone_type_properties:
            zone_type_properties += zt.properties
        for d in self._distancier_properties:
            distancier_properties += d.properties
        return self._basic_network_properties + zone_type_properties + distancier_properties
    
    def nb_properties(self):
        return len(self._basic_network_properties) + sum([zt.nb_properties() for zt in self._zone_type_properties]) + sum([d.nb_properties() for d in self._distancier_properties])
    
    def sql_basic_network_properties(self):
        list_sql = []
        if len (self._basic_network_properties) == 0:
            return "p.*"
        for prop in self._basic_network_properties:
            list_sql.append(f"{prop.to_sql(prefix='p')}")
        return list_sql
    
    def properties_sql(self):
        properties = []
        for prop in self._basic_network_properties:
            properties.append(prop.to_sql(prefix="p"))
        for zt in self._zone_type_properties:
            properties += zt.properties_sql()
        for d in self._distancier_properties:
            properties += d.properties_sql()
        return properties
    
    def join_sql(self):
        join = []
        for zt in self._zone_type_properties:
            join += ["JOIN ggo_zone AS z ON z.poi_id = p.poi_id", "JOIN ggo_zone_geounit AS zg ON zg.zone_id = z.zone_id"]
            join += zt.join_sql()
        for d in self._distancier_properties:
            j = 'poi_id_start' if d.distancier_session.network_id_start == self._network.network_id else 'poi_id_end'
            join += [f"JOIN ggo_distancier AS d ON d.{j} = p.poi_id", f"JOIN ggo_distancier_session AS ds ON ds.session_id = d.session_id"]
            join += d.join_sql()
        return join
    
    def where_sql(self):
        where = [f"p.network_id = {self._network.network_id}"]
        for zt in self._zone_type_properties:
            where += zt.where_sql()
        for d in self._distancier_properties:
            where += d.where_sql()
        return where
    
    def to_sql(self, idx=True, code=False, label=False):
        properties = ["p.poi_id AS poi_id"] if idx else []
        if code:
            properties.append("p.id AS poi_code")
        if label:
            properties.append("p.name AS poi_name")
        properties += self.properties_sql()
        group_by = idx + code + label + len(self._basic_network_properties)
        group_by = [str(i + 1) for i in range(group_by)]
        q = f"""
        SELECT
            {", ".join(properties)}
        FROM ggo_poi AS p
        {" ".join(self.join_sql())}
        WHERE {" AND ".join(self.where_sql())}
        GROUP BY {", ".join(group_by)}
        """
        return sqlparse.format(q, reindent=True, keyword_case='upper')
    


import pandas as pd
import geopandas as gpd
from shapely import wkt
from sqlalchemy import text

from .poi import Poi
from .properties.property import Property
from ..utils.types import pythonTypeToPostgresType

class Network:
    # Constructor
    def __init__(
            self,
            network_id:int,
            org_id:int,
            name:str,
            brand:str,
            network_type_id:str,
            is_default:bool,
            created_at:str,
            created_by:str,
            last_updated_at:str,
            last_updated_by:str,
            geolevel_id:int,
            description:str,
            org:'Org' # type: ignore
    ):
        # Infos
        self._network_id = network_id
        self._org_id = org_id
        self._name = name
        self._brand = brand
        self._network_type_id = network_type_id
        self._is_default = is_default
        self._created_at = created_at
        self._created_by = created_by
        self._last_updated_at = last_updated_at
        self._last_updated_by = last_updated_by
        self._geolevel_id = geolevel_id
        self._description = description
        self._org = org

    # Getters and setters
    @property
    def network_id(self): return self._network_id
    @property
    def org_id(self): return self._org_id
    @property
    def name(self): return self._name
    @property
    def brand(self): return self._brand
    @property
    def network_type_id(self): return self._network_type_id
    @property
    def is_default(self): return self._is_default
    @property
    def created_at(self): return self._created_at
    @property
    def created_by(self): return self._created_by
    @property
    def last_updated_at(self): return self._last_updated_at
    @property
    def last_updated_by(self): return self._last_updated_by
    @property
    def geolevel_id(self): return self._geolevel_id
    @property
    def description(self): return self._description
    @property
    def org(self): return self._org

    # Magic Methods
    def __str__(self):
        return f"Network({self._network_id}, {self._name}, {self._brand})"
    
    # Public Methods
    def number_of_pois(self)->int:
        query = text(f"SELECT COUNT(*) FROM ggo_poi WHERE network_id = {self._network_id}")
        with self._org.engine.connect() as conn:
            result = conn.execute(query)
            return result.scalar()
        
    def getGeoDataFrame(self)->gpd.GeoDataFrame:
        # Query
        query = f"SELECT * FROM ggo_poi WHERE network_id = {self._network_id}"
        # Get data from query
        gdf = gpd.read_postgis(query, self._org.engine, geom_col="geom")
        # return df
        return gdf

    def getPoisList(self)->gpd.GeoDataFrame:
        # Query
        query = f"SELECT * FROM ggo_poi WHERE network_id = {self._network_id}"
        # Get data from query
        gdf = gpd.read_postgis(query, self._org.engine, geom_col="geom")
        # return df
        return gdf
    
    def getPoiByCode(self, code:str)->Poi:
        # Query
        query = f"SELECT * FROM ggo_poi WHERE id = '{code}'"
        gdf = gpd.read_postgis(query, self._org.engine, geom_col="geom")
        # Data
        if len(gdf) > 0:
            data = gdf.iloc[0].to_dict()
            data.update({"org": self.org})
            return Poi(**data)
        else:
            raise Warning(f"Poi with code {code} not found in Network {self._name}")
        
    def getAllPois(self)->list:
        # Query
        query = f"SELECT * FROM ggo_poi WHERE network_id = {self._network_id}"
        gdf = gpd.read_postgis(query, self._org.engine, geom_col="geom")
        # Data
        pois = []
        for i in range(len(gdf)):
            data = gdf.iloc[i].to_dict()
            data.update({"org": self.org})
            pois.append(Poi(**data))
        return pois
    
    def getProperties(self, fast:int|None=None)->pd.DataFrame:
        # Basic properties
        # brand_property = Property(column="brand", json_info=None, dtype="TEXT")
        # Other properties
        query = f"""
            SELECT
                p.properties
            FROM ggo_poi AS p
            WHERE p.network_id = {self._network_id}
        """
        query += f""" LIMIT {fast}""" if fast else ""
        list_properties = self._org.query_df(query)['properties'].tolist()
        df_properties = pd.DataFrame(list_properties)
        df_prop = df_properties.dtypes.reset_index()
        df_prop.columns = ['columns', 'dtypes']
        df_prop['dtypes_postgres'] = df_prop['dtypes'].astype(str).apply(lambda x: pythonTypeToPostgresType(x))
        p = []
        for index, row in df_prop.iterrows():
            prop = Property(column='properties', dtype='JSONB', json_info={"key": row['columns'], "dtype": row['dtypes_postgres']})
            p.append(prop)
        return p
    
    @staticmethod
    def getAllNetworksFromZoneType(zone_type):
        # Query
        query = f"""
            SELECT
                DISTINCT n.*
            FROM ggo_zone AS z
            JOIN ggo_poi AS p ON p.poi_id = z.poi_id
            JOIN ggo_network AS n ON n.network_id = p.network_id
            WHERE z.zone_type_id = {zone_type.zone_type_id}
        """
        org = zone_type.org
        df = org.query_df(query)
        return [Network(**data, org=org) for data in df.to_dict(orient='records')]


from .geolevel_data_type import GeolevelDataType
from .geolevel import Geolevel
from .network import Network
from .poi import Poi
from .zone_geounit import ZoneGeounit
from .zone_type import ZoneType
from .zone import Zone
from .distancier_session import DistancierSession
from .cluster_session import ClusterSession

from .properties.property import Property
from .properties.network_properties import NetworkProperties
from .properties.zone_type_properties import ZoneTypeProperties
from .properties.geolevel_properties import GeolevelProperties
from .properties.geoleveldata_type_properties import GeoleveldataTypeProperties
from .properties.distancier_properties import DistancierProperties


from .property import Property
from ..distancier_session import DistancierSession
import sqlparse

class DistancierProperties:
    def __init__(
            self,
            distancier_session: DistancierSession,
            basic_distancier_properties: list[Property],
            geolevel_properties: list['GeolevelProperties'] | None = None, # type: ignore
            network_properties: list['NetworkProperties'] | None = None # type: ignore
            ):
        self._distancier_session = distancier_session
        self._distancier_properties = basic_distancier_properties
        self._geolevel_properties = geolevel_properties if geolevel_properties else []
        self._network_properties = network_properties if network_properties else []

    @property
    def distancier_session(self): return self._distancier_session
    @property
    def distancier_properties(self): return self._distancier_properties
    @property
    def geolevel_properties(self): return self._geolevel_properties
    @property
    def network_properties(self): return self._network_properties
    @property
    def properties(self):
        network_properties = []
        geolevel_properties = []
        for np in self._network_properties:
            network_properties += np.properties
        for gp in self._geolevel_properties:
            geolevel_properties += gp.properties
        return self._distancier_properties + network_properties + geolevel_properties
    
    def nb_properties(self):
        return len(self._distancier_properties) + sum([np.nb_properties() for np in self._network_properties]) + sum([gp.nb_properties() for gp in self._geolevel_properties])

    def sql_distancier_properties(self):
        list_sql = []
        if len(self._distancier_properties) == 0:
            return "d.*"
        for prop in self._distancier_properties:
            list_sql.append(f"{prop.to_sql(prefix='d')}")
        return ", ".join(list_sql)
    
    def properties_sql(self):
        properties = []
        for prop in self._distancier_properties:
            properties.append(prop.to_sql(prefix="d"))
        for np in self._network_properties:
            properties += np.properties_sql()
        for gp in self._geolevel_properties:
            properties += gp.properties_sql()
        return properties
    
    def join_sql(self, code=False):
        join = []
        if code:
            j = 'poi_id_start' if self._distancier_session.network_id_start else 'poi_id_end'
            join += [f"JOIN ggo_poi AS p ON p.poi_id = d.{j}"] if code else []
        for np in self._network_properties:
            j = 'poi_id_start' if np.network.network_id == self._distancier_session.network_id_start else 'poi_id_end'
            join += [f"JOIN ggo_poi AS p ON p.poi_id = d.{j}"] if code else []
            join += np.join_sql()
        for gp in self._geolevel_properties:
            j = 'geolevel_id_start' if gp.geolevel.geolevel_id == self._distancier_session.geolevel_id_start else 'geolevel_id_end'
            join += [f"JOIN ggo_geolevel AS g ON g.geolevel_id = ds.{j}"]
            join += gp.join_sql(source="d.geounit_code_end")
        return join
    
    def where_sql(self):
        where = [f"d.session_id = {self._distancier_session.session_id}"]
        for np in self._network_properties:
            where += np.where_sql()
        for gp in self._geolevel_properties:
            where += gp.where_sql()
        return where
    
    def to_sql(self, idx=True, code=False):
        properties = []
        if idx:
            properties += ["d.poi_id_start::INTEGER AS poi_id_start"] if self.distancier_session.network_id_start else []
            properties += ["d.geounit_code_start::TEXT AS geounit_code_start"] if self.distancier_session.geolevel_id_start else []
            properties += ["d.poi_id_end::INTEGER poi_id_end"] if self.distancier_session.network_id_end else []
            properties += ["d.geounit_code_end::TEXT AS geounit_code_end"] if self.distancier_session.geolevel_id_end else []
        if code:
            properties += ["p.id::TEXT AS poi_code_start"] if self.distancier_session.network_id_start else []
            properties += ["d.geounit_code_start::TEXT AS geounit_code_start"] if self.distancier_session.geolevel_id_start else []
            properties += ["p.id ::TEXT AS poi_code_end"] if self.distancier_session.network_id_end else []
            properties += ["d.geounit_code_end::TEXT AS geounit_code_end"] if self.distancier_session.geolevel_id_end else []
        properties += self.properties_sql()
        q = f"""
        SELECT
            {', '.join(properties)}
        FROM ggo_distancier AS d
        JOIN ggo_distancier_session AS ds ON ds.session_id = d.session_id
        {' '.join(self.join_sql(code=code))}
        WHERE {" AND ".join(self.where_sql())}
        """
        return sqlparse.format(q, reindent=True, keyword_case='upper')



PYTHON_POSTGRES_TYPE = {
    "int": "INTEGER",
    "int64": "BIGINT",
    "float": "FLOAT",
    "float64": "DOUBLE PRECISION",
    "str": "TEXT",
    "bool": "BOOLEAN",
    "datetime": "TIMESTAMP",
    "date": "DATE",
    "time": "TIME",
    "dict": "JSON",
    "list": "ARRAY",
    "NoneType": "NULL",
    "object": "TEXT",
    "geometry": "GEOMETRY",
    "nan": "FLOAT",
}

def pythonTypeToPostgresType(python_type: str) -> str:
    """
    Convert a Python type to a PostgreSQL type.

    Args:
        python_type (str): The Python type as a string.

    Returns:
        str: The corresponding PostgreSQL type.
    """
    if python_type.lower() in PYTHON_POSTGRES_TYPE:
        return PYTHON_POSTGRES_TYPE[python_type.lower()]
    else:
        raise ValueError(f"Unsupported Python type: {python_type.lower()}")


from ..geolevel_data_type import GeolevelDataType
from .property import Property

class GeoleveldataTypeProperties:
    def __init__(self, geoleveldata_type : GeolevelDataType, geoleveldata_properties: list[Property]):
        self._geoleveldata_type = geoleveldata_type
        self._geoleveldata_properties = geoleveldata_properties

    @property
    def geoleveldata_type(self): return self._geoleveldata_type
    @property
    def geoleveldata_properties(self): return self._geoleveldata_properties
    @property
    def properties(self): return self._geoleveldata_properties

    def nb_properties(self):
        return len(self._geoleveldata_properties)

    def sql_geoleveldata_properties(self):
        list_sql = []
        if len(self._geoleveldata_properties) == 0:
            return "gd.*"
        for prop in self._geoleveldata_properties:
            list_sql.append(f"{prop.to_sql(prefix='gd')}")
        return ", ".join(list_sql)
    
    def properties_sql(self):
        properties = []
        for prop in self._geoleveldata_properties:
            properties.append(prop.to_sql(prefix="gd"))
        return properties
    
    def join_sql(self):
        join = []
        return join
    
    def where_sql(self):
        where = [f"gd.geoleveldata_type_id = {self._geoleveldata_type.geoleveldata_type_id}"]
        return where
    
    def to_sql(self):
        return ""


from ..geolevel import Geolevel
import sqlparse

class GeolevelProperties:
    # Constructor
    def __init__(self, geolevel: Geolevel, geoleveldata_type_properties: list['GeoleveldataTypeProperties']): # type: ignore
        self._geolevel = geolevel
        self._geoleveldata_type_properties = geoleveldata_type_properties if geoleveldata_type_properties else []

    # Getters and setters
    @property
    def geolevel(self): return self._geolevel
    @property
    def geoleveldata_type_properties(self): return self._geoleveldata_type_properties
    @property
    def properties(self):
        geoleveldata_type_properties = []
        for gdtp in self._geoleveldata_type_properties:
            geoleveldata_type_properties += gdtp.properties
        return geoleveldata_type_properties
    
    def nb_properties(self):
        return len(self._geoleveldata_type_properties) + sum([gdtp.nb_properties() for gdtp in self._geoleveldata_type_properties])

    def sql_geoleveldata_type_properties(self):
        list_sql = []
        if len(self._geoleveldata_type_properties) == 0:
            return "gd.*"
        for prop in self._geoleveldata_type_properties:
            list_sql.append(f"{prop.to_sql(prefix='gd')}")
        return ", ".join(list_sql)
    
    def properties_sql(self):
        properties = []
        for gdtp in self._geoleveldata_type_properties:
            properties += gdtp.properties_sql()
        return properties
    
    def join_sql(self, source="zg.geounit_code"):
        join = []
        for gdtp in self._geoleveldata_type_properties:
            join += [f"JOIN ggo_geoleveldata AS gd ON gd.geounit_code = {source}"]
            join += gdtp.join_sql()
        return join
    
    def where_sql(self):
        where = [f"g.geolevel_id = {self._geolevel.geolevel_id}"]
        for gdtp in self._geoleveldata_type_properties:
            where += gdtp.where_sql()
        return where
    
    def to_sql(self, code=True, source="gd"):
        properties = [f"{source}.geounit_code AS geounit_code"] if code else []
        properties += self.properties_sql()
        q = f"""
        SELECT
            {', '.join(properties)}
        FROM ggo_geolevel AS g
        {' '.join(self.join_sql(source=source))}
        WHERE {" AND ".join(self.where_sql())}
        """
        return sqlparse.format(q, reindent=True, keyword_case='upper')